#include <avr/pgmspace.h>
#include <util/delay.h>
#include "dogm.h"

#ifndef LCD_PORT_WRITE
	#error "No WRITE Port Specified!"
#endif
#ifndef LCD_PORT_READ
	#error "No READ Port Specified!"
#endif
#ifndef LCD_PORT_DIR
	#error "No DDR Port Specified!"
#endif
#ifndef WRAP_LINES
	#error "No Wrapping Method Specified (WRAP_LINES 0/1)!"
#endif

#if LCD_TYPE==81
static uint8_t _base_y[1]={0x80};
uint8_t _lcd_maxx=80;
#elif LCD_TYPE==162
static uint8_t _base_y[2]={0x80,0xC0};
uint8_t _lcd_maxx=40;
#elif LCD_TYPE==163
static uint8_t _base_y[3]={0x80,0x90,0xa0};
uint8_t _lcd_maxx=16;
#else
	#error "Type Of LCD Not Specified (LCD_TYPE 81/162/163)!"
#endif

uint8_t _lcd_x,_lcd_y;

void _dog_checkbusy(void)
	{
	uint8_t data;

	// D7-D4 auf Eingang schalten:
	LCD_PORT_DIR = 0x0F;
	data = 0xFF;

	// D7: BusyFlag = 0: not busy
	while (bit_is_set (data,7))			// solange D7 gesetzt ist, nicht weitermachen
		{
		// RS low, RW high: High-Nibble des Display-Bytes mit Adresse und BusyFlag lesen:
		LCD_PORT_WRITE = 0x06;				//RS0 RW1 E1
		_delay_us(1);

		data = LCD_PORT_READ;	 			// 1. Nibble (MSB)

		LCD_PORT_WRITE &= ~(1<<E);		    // E0
		_delay_us(1);

		// low nibble verwerfen:
		LCD_PORT_WRITE |= (1 << E);		// E1
		_delay_us(1);
		
		LCD_PORT_WRITE &= ~(1<<E);		    // E0
		}
	// D7-D4 auf Ausgang schalten:
	LCD_PORT_DIR =0xFF;
	}


uint8_t _dog_readbusy (void)
	{
	uint8_t data;

	//Port Direction Register: data=in, control=out
	LCD_PORT_DIR = 0x0F;

	LCD_PORT_WRITE = 0x06;									//RS0 RW1 E1
	_delay_us(1);

	data = LCD_PORT_READ;	 								//1. Nibble (MSB)

	LCD_PORT_WRITE &=~ (1<<E);								//E 0
	_delay_us(1);

	LCD_PORT_WRITE = 0x06;									//RS0 RW1 E1
	_delay_us(1);

	LCD_PORT_WRITE &=~ (1<<E);								//E 0

	//Port Direction Register: data+control=out
	LCD_PORT_DIR = 0xFF;

	return(data);
	}

//
//write a data byte with busy-check
void _dog_write_data (uint8_t data)
	{
	//check if busy bit is clear
	_dog_checkbusy();												//LCD ready?

	//Set Port Direction Register to Output for data and control bus
	LCD_PORT_DIR=0xFF;							//DDRD output

	LCD_PORT_WRITE = (data&0xF0) | 0x05;					//1. Nibble + RS1 RW0 E1
	_delay_us(1);

	LCD_PORT_WRITE &=~ (1<<E);					//E 0
	_delay_us(1);

	LCD_PORT_WRITE = ((data&0x0F)<<4) | 0x05;	//2. Nibble + RS1 RW0 E1
	_delay_us(1);

	LCD_PORT_WRITE &=~ (1<<E);					//E 0
	}

//
//write an instruction byte with busy-check
void _dog_write_inst (uint8_t data)
	{
	//check if busy bit is clear
	_dog_checkbusy();												//LCD ready?

	//Set Port Direction Register to Output for data and control bus
	LCD_PORT_DIR=0xFF;							//DDRD output

	LCD_PORT_WRITE = (data&0xF0) | 0x04;		//1. Nibble + RS0 RW0 E1
	_delay_us(1);

	LCD_PORT_WRITE &=~ (1<<E);					//E 0
	_delay_us(1);

	LCD_PORT_WRITE = ((data&0x0F)<<4) | 0x04;	//2. Nibble + RS0 RW0 E1
	_delay_us(1);

	LCD_PORT_WRITE &=~ (1<<E);					//E 0
	}

//
//write an instruction byte without busy-check
//(required for the first 3 init bytes according to datasheet)
void _dog_write_nocheck (uint8_t data)
	{
	//Set Port Direction Register to Output for data and control bus
	LCD_PORT_DIR=0xFF;							//DDRD output

	LCD_PORT_WRITE = (data&0xF0) | 0x04;		//1. Nibble + RS0 RW0 E1
	_delay_us(1);

	LCD_PORT_WRITE &=~ (1<<E);					//E 0
	_delay_us(1);

	LCD_PORT_WRITE = ((data&0x0F)<<4) | 0x04;	//2. Nibble + RS0 RW0 E1
	_delay_us(1);

	LCD_PORT_WRITE &=~ (1<<E);					//E 0
	}

//
//Display character at current cursor position
//Input:    character
void _dog_putc(uint8_t c)
	{
    uint8_t pos;
    pos = _dog_readbusy();   	// read busy-flag and address counter
    if (c=='\n')
		{
//        	lcd_newline(pos);	// muss erst noch angepasst werden...
		}
    else
		{
		#if WRAP_LINES==1
			#if LCD_TYPE==81
				if (pos==0x07)
					{
					_dog_write_inst((1<<DDRAM)+0x00);
					}
			#elif LCD_TYPE==162
				if (pos==0x0f)
					{
					_dog_write_inst((1<<DDRAM)+0x40);
					}
				else if (pos==0x4f)
					{
					_dog_write_inst((1<<DDRAM)+0x00);
					}
			#elif LCD_TYPE==163
				if (pos==0x0f)
					{
					_dog_write_inst((1<<DDRAM)+0x10);
					}
				else if (pos==0x1f)
					{
					_dog_write_inst((1<<DDRAM)+0x20);
					}
				else if (pos==0x2f)
					{
					_dog_write_inst((1<<DDRAM)+0x00);
					}
			#endif
			_dog_readbusy();
		#endif
		_dog_write_data(c);
		}
	}

void dog_init (void)
	{
	//Wait 40ms
	_delay_ms(40);
	
	_dog_write_nocheck(0x30);
	_delay_ms(2);
	_dog_write_nocheck(0x30);
	_delay_us(30);
	_dog_write_nocheck(0x30);
	_delay_us(30);

	//check if busy bit is clear
	_dog_checkbusy();												//LCD ready?

	_dog_write_inst(0x20);
	_delay_us(30);
	_dog_write_inst(0x29);
	_delay_us(30);
	_dog_write_inst(0x14);
	_delay_us(30);
	_dog_write_inst(0x78);
	_delay_us(30);
	_dog_write_inst(0x5e);
	_delay_us(30);
	_dog_write_inst(0x6a);
	_delay_us(30);
	_dog_write_inst(0x0c);
	_delay_us(30);
	_dog_write_inst(0x01);
	_delay_us(30);
	_dog_write_inst(0x06);
	_delay_us(30);
	}

void dog_clear(void)
	{
	_dog_write_inst(0x02);   	// cursor home
	_delay_us(30);
	_dog_write_inst(0x0c); 		// cursor off
	_delay_us(30);
	_dog_write_inst(0x01);   	// clear
	_delay_us(30);
	}

//set LCD display position x=0..39 y=0..3
void dog_gotoxy(uint8_t x, uint8_t y)
	{
	_dog_write_inst(_base_y[y]+x);
	_lcd_x=x;
	_lcd_y=y;
	}

//
//Display string without auto linefeed
//Input:    string
void dog_puts(const char *s)
	{
    register char c;

    while ((c=*s++))
		{
        _dog_putc(c);
		}
	}

//
//Display string from program memory (flash) without auto linefeed
//Input:     string from program memory
void dog_putsf(const char *progmem_s)
	{
    register char c;

    while ((c=pgm_read_byte(progmem_s++)))
		{
        _dog_putc(c);
		}
	}

void dog_contrast (unsigned short contrast)
	{
	if ((contrast>=0x00)&&(contrast<=0x0f))
		{
        _dog_write_inst(0x29);
        _dog_write_inst(0x70|contrast);
        _dog_write_inst(0x51);
		}
	}
